/*

    File: Mesh.cpp
	Copyright (C) 2000-2007 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#include <windows.h>
#include <gl/gl.h>
#include <assert.h>
#include <stdio.h>

#include "../m_foundation/main.h"
#include "../m_foundation/const.h"

#include "mesh.h"
#include "main.h"
#include "t3d.h"	//t3dvertex

CFUNC CFunc_saveMesh;

static int useTexture = 1;
static int useBlend = 1;
static quat_t light0position = {30, 0, 30, 1};
static quat_t bgcolor = {0,0,0,1};

static CVarw *local_cvars[] = {
	new CVarw(&useTexture,		"useTexture"),
	new CVarw(&useBlend,		"useBlend"),
	new CVarw(&light0position, "lightpos"),
	new CVarw(&bgcolor,		"bgcolor"),
	new CVarw(0,				"savemesh",		CFunc_saveMesh),
};

class TriBuffer {
	int tri[65536][3];
	int count;
public:
	TriBuffer() {
		init();
	}

	void init(void) {
		memset(tri, 0, sizeof(tri));
		count = 0;
	}

	void addTri(int a, int b, int c) {
		if (count >= numberof(tri)) {
			main_console->Print("tri buffer full\n");
			return;
		}
		tri[count][0] = a;
		tri[count][1] = b;
		tri[count][2] = c;
		count++;
	}

	int getTriVtxIndex(int tri, int corner) {
		return this->tri[tri][corner];
	}

	int getCount() { return count; }
};

typedef struct {
	vector_t x;
	vector_t n;
	float st[2];
} vertex_t;

class VertexBuffer {
private:
	vertex_t vertex[65536];
	int count;

public:

	VertexBuffer() {
		init();
	}

	void init(void) {
		memset(vertex, 0, sizeof(vertex));
		count = 0;
	}

	//returns the first vertex of a set of vertices
	int alloc(int size) {
		if ((int)numberof(vertex) - count < size) {
			main_console->Print("no more room left!\n");
			return 0;
		}
		count += size;
		return count - size;
	}

	void setVtx(int index, vector_t v) {
		if (index < 0 || index >= count) return;
		memset(&vertex[index], 0, sizeof(vertex_t));
		vertex[index].x = v;
	}

	void setTexCoord(int index, float s, float t) {
		if (index < 0 || index >= count) return;
		vertex[index].st[0] = s;
		vertex[index].st[1] = t;
	}

	const float *getVertexPtr(int index) {
		return vertex[index].x.v;
	}

	const float *getNormalPtr(int index) {
		return vertex[index].n.v;
	}

	const float *getTexCoordPtr(int index ) {
		return vertex[index].st;
	}

	int getCount() { return count; }

	void genNormals(TriBuffer *tri) {
		assert(tri);

		for (int i = 0; i < count; i++) {
			vertex[i].n = vectorZero;
		}

		for (i = 0; i < tri->getCount(); i++) {

			vector_t n;

			VectorPlaneUnitNormal(
				&vertex[tri->getTriVtxIndex(i,2)].x,
				&vertex[tri->getTriVtxIndex(i,1)].x,
				&vertex[tri->getTriVtxIndex(i,0)].x, &n);

			VectorAdd(&vertex[tri->getTriVtxIndex(i,0)].n, &n, &vertex[tri->getTriVtxIndex(i,0)].n);
			VectorAdd(&vertex[tri->getTriVtxIndex(i,1)].n, &n, &vertex[tri->getTriVtxIndex(i,1)].n);
			VectorAdd(&vertex[tri->getTriVtxIndex(i,2)].n, &n, &vertex[tri->getTriVtxIndex(i,2)].n);
		}

		for (i = 0; i < count; i++) {
			VectorUnit(&vertex[i].n, &vertex[i].n);
		}

	}
};



//app-spec.
Texture texture;
VertexBuffer vertexBuffer;
TriBuffer triBuffer;

void setRelVtx(int index, float x, float y, float z, float s, float t) {
	if (!T3DTestStackBounds()) return;

	vector_t v = {x,y,z};
	T3DVertex(&v, &v);
	vertexBuffer.setVtx(index, v);
	vertexBuffer.setTexCoord(index, s, t);
}

class VertexRange {
public://private:
	int start;
	int size;

//as much as I hate global variables, thats what this is becoming...
	VertexBuffer *vtx;

public:
	//by default reference the single static variable
	//-- this is going to become our main ctor method
	VertexRange() {
		this->vtx = &vertexBuffer;
		this->start = 0;
		this->size = 0;
	}

	VertexRange(VertexBuffer *vtx) {
		this->vtx = vtx;
		this->start = 0;
		this->size = 0;
	}

	void alloc(int size) {
		this->size = size;
		this->start = vtx->alloc(size);
	}

	void setRelVtx(int index, float x, float y, float z, float s, float t) {
		::setRelVtx(start + index, x, y, z, s, t);
	}

	void bind(VertexRange *other, TriBuffer *tri) {
		
		//write poly between the vertices

		for (int i = 0; i < size; i++) {
			int i2 = i;
			if (i2 >= other->size) i2 = other->size - 1;
			tri->addTri(
				start + i,
				start + ((i+1) % size),
				other->start + i2
			);
		}

		for (i = 0; i < other->size; i++) {
			int i2 = i+1;
			if (i2 > size) i2 = size;
			if (i2 == size) i2 = 0;
			tri->addTri(
				start + i2,
				other->start + ((i+1) % other->size),
				other->start + i
			);
		}
	}
};

char mesh_texture_filename[256] = {0};

static bool meshSave(const char *filename) {

	char obj_filename[256], mtl_filename[256];
	sprintf(obj_filename, "%s.obj", filename);
	sprintf(mtl_filename, "%s.mtl", filename);

	FILE *file = fopen(obj_filename, "w");
	if (!file) return false;

	for (int i = 0; i < vertexBuffer.getCount(); i++) {
		const float *v = vertexBuffer.getVertexPtr(i);
		fprintf(file, "v %f %f %f\n", v[0], v[1], v[2]);
	}

	for (i = 0; i < vertexBuffer.getCount(); i++) {
		const float *v = vertexBuffer.getTexCoordPtr(i);
		fprintf(file, "vt %f %f\n", v[0], v[1]);
	}

	for (i = 0; i < vertexBuffer.getCount(); i++) {
		const float *v = vertexBuffer.getNormalPtr(i);
		fprintf(file, "vn %f %f %f\n", v[0], v[1], v[2]);
	}

	fprintf(file, "mtllib %s\n", mtl_filename);
	fprintf(file, "g g0\n");
	fprintf(file, "usemtl tex\n");	//where "tex" is the object texture (expand support for multiple textures later

	for (i = 0; i < triBuffer.getCount(); i++) {
		fprintf(file, "f ");
		for (int j = 0; j < 3; j++) {
			int index = triBuffer.getTriVtxIndex(i,j) + 1;	//one is the base index for obj files
			fprintf(file, "%d/%d/%d ", index, index, index);
		}
		fprintf(file, "\n");
	}

	fclose(file);

	file = fopen(mtl_filename, "w");
	if (file) {
		fprintf(file, "newmtl tex\n");	//where "tex" is the object texture
		fprintf(file, "Ka 1 1 1\n");
		fprintf(file, "Kd 1 1 1\n");
		fprintf(file, "Ks 1 1 1\n");
		fprintf(file, "illum 1\n");
		fprintf(file, "Ns 1\n");
		if (mesh_texture_filename[0]) {
			fprintf(file, "map_Kd %s\n", mesh_texture_filename);
		}
		fclose(file);
	}

	return true;
}

void CFunc_saveMesh(int argc, char **argv) {
	if (argc < 2) {
		main_console->Print("usage %s <output-filename>\n", argv[0]);
	} else {
		meshSave(argv[1]);
	}
}

// **************** SCRIPT INTERFACE ****************

typedef union vtxrange_u {
	unsigned int l;
	struct {
		unsigned short start;
		unsigned short size;
	};
} vtxrange_t;

//allocs from the main vertexBuffer object
int vertexRangeAlloc(int size) {
	vtxrange_t vr;
	vr.size = size;
	vr.start = vertexBuffer.alloc(size);
	return vr.l;
}

//binds to the main triBuffer objects
void BindRange(int l1, int l2) {
	vtxrange_t *v1 = (vtxrange_t *)(&l1);
	vtxrange_t *v2 = (vtxrange_t *)(&l2);

	//write poly between the vertices

	for (int i = 0; i < v1->size; i++) {
		int i2 = i;
		if (i2 >= v2->size) i2 = v2->size - 1;
		triBuffer.addTri(
			v1->start + i,
			v1->start + ((i+1) % v1->size),
			v2->start + i2
		);
	}

	for (i = 0; i < v2->size; i++) {
		int i2 = i+1;
		if (i2 > v1->size) i2 = v1->size;
		if (i2 == v1->size) i2 = 0;
		triBuffer.addTri(
			v1->start + i2,
			v2->start + ((i+1) % v2->size),
			v2->start + i
		);
	}
}

void setRangeVtx(int range, int index, float x, float y, float z, float s, float t) {
	vtxrange_t *v = (vtxrange_t *)(&range);
	setRelVtx(v->start + index, x, y, z, s, t);
}

void meshSetTexture(const char *filename) {

	strncpy(mesh_texture_filename, filename, sizeof(mesh_texture_filename)-1);
	texture.Create( filename );
}

// **************** SYSTEM INTERFACE ****************

void meshRender(void) {

	glClearColor(bgcolor.x, bgcolor.y, bgcolor.z, bgcolor.w);

	glLightfv(GL_LIGHT0, GL_POSITION, light0position.v);

	if (useTexture) {
		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D, texture.texture_id);
	} else {
		glDisable(GL_TEXTURE_2D);
	}

	if (useBlend) {
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glEnable(GL_BLEND);
	} else {
		glDisable(GL_BLEND);
	}

	glBegin(GL_TRIANGLES);
	for (int i = 0; i < triBuffer.getCount(); i++) {
		for (int j = 0; j < 3; j++) {
			int index = triBuffer.getTriVtxIndex(i,j);
			glTexCoord2fv(vertexBuffer.getTexCoordPtr(index));
			glNormal3fv(vertexBuffer.getNormalPtr(index));
			glVertex3fv(vertexBuffer.getVertexPtr(index));
		}
	}
	glEnd();

//	glDisable(GL_TEXTURE_2D);
//	glColor3f(1,1,1);
//	glBegin(GL_POINTS);
//	for (i = 0; i < vertexBuffer.getCount(); i++) {
//		glVertex3fv(vertexBuffer.getVertexPtr(i));
//	}
//	glEnd();
}

void meshReset(void) {
	//init storage:
	vertexBuffer.init();
	triBuffer.init();

	memset(mesh_texture_filename,0,sizeof(mesh_texture_filename));
}

void meshPostLoad(void) {

	vertexBuffer.genNormals(&triBuffer);
	
	main_console->Print("created %d vtxs, %d tris\n",
		vertexBuffer.getCount(), triBuffer.getCount());

	box_t box = {0,0,0,0,0,0};
	for (int i = 0; i < vertexBuffer.getCount(); i++) {
		vector_t *v = (vector_t *)vertexBuffer.getVertexPtr(i);
		if (!i) {
			box.min = box.max = *v;
		} else {
			BoxStretch(&box, v);
		}
	}
	main_console->Print("min range: %f %f %f\n", box.min.x, box.min.y, box.min.z);
	main_console->Print("max range: %f %f %f\n", box.max.x, box.max.y, box.max.z);
}
