/*

    File: T3D.cpp
	Copyright (C) 2000-2007 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#include "../m_math3d/geometry.h"
#include "../m_foundation/main.h"	//console
#include "main.h"
#include "t3d.h"

#undef UNIFORM_T3D

#ifdef UNIFORM_T3D
typedef struct {
	quat_t		r;	//rotation
	scalar_t	s;	//scale
	vector_t	t;	//translation
} t3d_t;	//3D transform
#else
typedef struct {
	float m[4][4];	//m[column][row]
} t3d_t;
#endif

static t3d_t t3dIdentity;

static t3d_t t3dStack[256];
static int t3dStackIndex = 0;

void T3DStackInit(void) {
	t3dStackIndex = 0;
	t3dStack[t3dStackIndex] = t3dIdentity;

	//just in case . . .
#ifdef UNIFORM_T3D
	t3dIdentity.r = quatIdentity;
	t3dIdentity.s = 1.f;
	t3dIdentity.t = vectorZero;
#else
	t3dIdentity.m[0][0] = t3dIdentity.m[1][1] = 
		t3dIdentity.m[2][2] = t3dIdentity.m[3][3] = 1.f;
#endif

}

bool T3DTestStackBounds(void) {
	if (t3dStackIndex < 0) {
		main_console->Print("T3D Stack Underflow!!!\n");
		return false;
	}
	if (t3dStackIndex >= numberof(t3dStack)) {
		main_console->Print("T3D Stack Overflow!!!\n");
		return false;
	}
	return true;
}

void T3DPush(void) {
	t3dStackIndex++;
	if (!T3DTestStackBounds()) return;
	t3dStack[t3dStackIndex] = t3dStack[t3dStackIndex-1];
}

void T3DPop(void) {
	t3dStackIndex--;
	T3DTestStackBounds();
}

t3d_t *T3DGetCurrent(void) {
	int index = T3DTestStackBounds() ? t3dStackIndex : numberof(t3dStack) - 1;
	return &t3dStack[index];
}

void T3DMul(const t3d_t *a, const t3d_t *b, t3d_t *res) {
	t3d_t c;
#ifdef UNIFORM_T3D
	//calc new translation
	VectorScale(&b->t, a->s, &c.t);
	QuatRotateVector(&c.t, &a->r, &c.t);
	VectorAdd(&c.t, &a->t, &c.t);
	//calc new scale
	c.s = a->s * b->s;
	//calc new rotation
	QuatMul(&a->r, &b->r, &c.r);
	//copy into result
#else
	for (int i = 0; i < 4;i ++) {
		for (int j = 0; j < 4; j++) {
			c.m[i][j] = 0;
			for (int k = 0; k < 4; k++) {
				c.m[i][j] += a->m[k][j] * b->m[i][k];
			}
		}
	}
#endif
	*res = c;
}

void T3DScale(float s) {
	t3d_t *t = T3DGetCurrent();
	t3d_t r = t3dIdentity;
#ifdef UNIFORM_T3D
	r.s = s;
#else
	r.m[0][0] = r.m[1][1] = r.m[2][2] = s;
#endif
	T3DMul(t, &r, t);
}

void T3DTranslate(float x, float y, float z) {
	t3d_t *t = T3DGetCurrent();
	t3d_t r = t3dIdentity;
#ifdef UNIFORM_T3D
	r.t.x = x;
	r.t.y = y;
	r.t.z = z;
#else
	r.m[3][0] = x;
	r.m[3][1] = y;
	r.m[3][2] = z;
#endif
	T3DMul(t, &r, t);
}

void T3DRotate(float degrees, float x, float y, float z) {
	t3d_t *t = T3DGetCurrent();
	t3d_t r = t3dIdentity;
#ifdef UNIFORM_T3D
	r.r.x = x;
	r.r.y = y;
	r.r.z = z;
	r.r.w = degrees;
	VectorUnit(&r.r.vec, &r.r.vec);
	AngleAxisToQuat(&r.r, &r.r);
#else
	vector_t a = {x,y,z};
	VectorUnit(&a, &a);
	matrix_t m;
	MatrixRotation(degrees * (float)M_PI / 180.f, a.x, a.y, a.z, &m);
	for (int i = 0; i < 3; i++) {
		for (int j = 0; j < 3; j++) {
			r.m[i][j] = m.v[i][j];
		}
	}
#endif
	T3DMul(t, &r, t);
}

////non-opengl -- override the scale value:
//void T3DSetScale(float scale) {
//	t3d_t *t = T3DGetCurrent();
//	t->s = scale;
//}

void T3DVertex(const vector_t *b, vector_t *res) {
	vector_t c;
	t3d_t *a = T3DGetCurrent();
#ifdef UNIFORM_T3D
	VectorScale(b, a->s, &c);
	QuatRotateVector(&c, &a->r, &c);
	VectorAdd(&c, &a->t, &c);
#else
	for (int i = 0; i < 3; i++) {
		c.v[i] = b->x * a->m[0][i] + b->y * a->m[1][i] + b->z * a->m[2][i] + a->m[3][i];
	}
#endif
	*res = c;
}
